/************************************************************************/
/* astronomy.c - Global source file for astronomy stuff.		*/
/************************************************************************/
/*
 *  (C)2000 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/
/* History:								*/
/* (12-09-2000) Original creation date. Todd Shadburn			*/
/************************************************************************/

#include <math.h>
#include <time.h>
#include <errno.h>

#include <stdlib.h>
#include <stdio.h>

#include "astronomy.h"


double	rev( double x )
{
	return x - floor(x/360.0)*360.0;
}


double	cbrt( double x )
{
	if ( x > 0.0 )
		return exp( log(x) / 3.0 );
	else if ( x < 0.0 )
		return -cbrt(-x);
	else
		return 0.0;
}


double get_daynumber( int year, int month, int day, double ut)
{
	double v1, v2, v3;
	v1 = 367.0*year;
	v2 = floor((7.0*(year+floor((month+9.0)/12.0)))/4.0);
	v3 = floor(275.0*month/9.0);
	return (v1 - v2 + v3 + day - 730530.0 + (ut/24.0));
}


double get_lst_degrees(double daynumber, double ut, double longitude)
{
	return (98.9818 + 0.985647352 * daynumber + ut * 15.0 + longitude);
}


double get_hour_angle(double lst, double ra)
{
	return (lst - ra);
}


double get_altitude_from_dec(double latitude, double dec, double ha)
{
	return (sin(sin(latitude) * sin(dec) + cos(latitude) *
		cos(dec) * cos(ha)));
}


double get_azimuth_from_ra_and_dec(double latitude, double dec, double ha)
{
	return (atan2(sin(ha), (cos(ha)*sin(latitude)-tan(dec)*cos(latitude))));
}


double get_degrees_from_hms(double hour, double minute, double sec)
{
	return (((double)hour + (double)minute/60.0 + (double)sec/3600.0) * 15.0);
}


int get_hms_from_degrees(double degrees, double *hour, double *minute, double *sec)
{
	double remander;	
	remander = degrees/15.0;
	*hour = floor(remander);
	remander = (remander - *hour)*60.0;
	*minute = floor(remander);
	remander = (remander - *minute)*60.0;
	*sec = remander;
	return 0;
}


double get_hours_from_hms(double hour, double minute, double sec)
{
	return (hour + minute/60.0 + sec/3600.0);
}


int get_dms_from_degrees(double degrees, double *deg, double *minute, double *second)
{
	double f;
	*deg = floor(degrees);
	f = (degrees - *deg) * 60.0;
	*minute = floor(f);
	*second = (f - *minute) * 60.0;
	return 0;
}


double normalize_hour(double hour)
{
	return (hour - (floor(hour/24.0)*24.0));
}
	
double	normalize_degree( double deg )
{
	return (deg - (floor(deg/360.0)*360.0));
}

int coord_spherical_to_rectangular(sphercoord_t *sc, rectcoord_t *rc)
{
	if(!sc || !rc)
		return -1;
	rc->x = sc->r * cos(sc->ra*DEGRAD) * cos(sc->dec*DEGRAD);
	rc->y = sc->r * sin(sc->ra*DEGRAD) * cos(sc->dec*DEGRAD);
	rc->z = sc->r * sin(sc->dec*DEGRAD);
	return 0 ;
}


int coord_rectangular_to_spherical(rectcoord_t *rc, sphercoord_t *sc)
{
	if(!rc || !sc)
		return -1;
	sc->r = sqrt(rc->x * rc->x + rc->y * rc->y + rc->z * rc->z);
	if(rc->x == 0.0 && rc->y == 0.0)
		sc->ra = 0.0;
	else
		sc->ra = atan2(rc->y, rc->x);
	sc->dec = atan2(2, sqrt(rc->x * rc->x + rc->y * rc->y));
	return 0;
}


double get_current_lst(int timezonediff, double longitude)
{
	struct tm *tmtime;
	time_t tsec;
	double ut;
	double degree;
	double daynumber;
	tsec = time(NULL);
	tmtime = localtime(&tsec);
	ut = get_current_ut(timezonediff);
	ut += timezonediff;
	if(tmtime->tm_isdst)
		ut -= 1.0;
	ut = normalize_hour(ut);
	daynumber = get_daynumber((tmtime->tm_year+1900.0), (tmtime->tm_mon+1.0),
		tmtime->tm_mday, ut);
	degree = get_lst_degrees(daynumber, ut, longitude);
	degree = rev(degree);
	return degree;
}


double get_current_ut(int timezonediff)
{
	struct tm *tmtime;
	time_t tsec;
	double ut;
	tsec = time(NULL);
	tmtime = localtime(&tsec);
	ut = tmtime->tm_hour + (double)tmtime->tm_min/60.0 + (double)tmtime->tm_sec/3600.0;
	ut = normalize_hour(ut + (-1*timezonediff));
	return ut;
}



/************************************************************************/
/*
 * Revamp on 2004-03-08
*/
/************************************************************************/

double normalize (double value, double normal)
{
	return (value - (floor(value/normal)*normal));
}
	

double current_ut_hours (int utcoffset)
{
	time_t t;
	double hours;
	DateTime dt;
	datetime_get_ut_current (&dt, utcoffset);
	hours = dt.hour + (dt.minute/60) + (dt.second/3600);
	hours = normalize (hours, 24.0);
	return hours;
}

/* The DateTime argument should be UTC */
double epoch_j2000_days (DateTime *dt)
{
	double days;
	days = 367.0 * dt->year - floor((7.0/4.0) *
		(dt->year+floor((dt->month)/12.0))) +
		floor(275.0 * dt->month / 9.0) +
		(double)dt->day - 730531.5 +
		(hours_from_datetime(dt)/24.0);
	return days;
}

double epoch_j2000_centuries (double epochdays)
{
	double centuries;
	centuries = epochdays / 36525.0;
	return centuries;
}


int datetime_get_local_current (DateTime *dt)
{
	time_t t;
	struct tm tmtime;
	if (dt == NULL)
		return EINVAL;
	t = time(NULL);
	localtime_r (&t, &tmtime);
	dt->year = tmtime.tm_year+1900;
	dt->month = tmtime.tm_mon+1;
	dt->day = tmtime.tm_mday;
	dt->hour = tmtime.tm_hour;
	dt->minute = tmtime.tm_min;
	dt->second = tmtime.tm_sec;
	return 0;
}


int datetime_get_ut_current (DateTime *dt, int utcoffset)
{
	int days;
	time_t t;
	struct tm tmtime;
	if (dt == NULL)
		return EINVAL;
	t = time(NULL);
	localtime_r (&t, &tmtime);
	dt->year = tmtime.tm_year+1900;
	dt->month = tmtime.tm_mon+1;
	dt->day = tmtime.tm_mday;
	dt->hour = tmtime.tm_hour;
	dt->minute = tmtime.tm_min;
	dt->second = tmtime.tm_sec;
	dt->hour += (utcoffset * -1);
	if (dt->hour > 23) {
		dt->hour = normalize (dt->hour, 24);
		dt->day++;
	} else if (dt->hour < 0) {
		dt->hour = normalize (dt->hour, 24);
		dt->day--;
	}
	days = days_in_month (dt->month, dt->year);
	if (dt->day > days) {
		dt->month++;
		dt->day = 1;
	} else if (dt->day < 1) {
		dt->month--;
		dt->day = days_in_month (dt->month, dt->year);
	}
	if (dt->month > 12) {
		dt->year++;
		dt->month = 1;
	} else if (dt->month < 1) {
		dt->year--;
		dt->month = 12;
	}
	return 0;
}


double hours_from_datetime (DateTime *dt)
{
	double hours;
	if (dt == NULL)
	return EINVAL;
	hours = (double)dt->hour + (dt->minute/60.0) + (dt->second/3600.0);
	return hours;
}


int days_in_month (int month, int year)
{
	int days;
	if (month < 1 || month > 12) {
		/* what calendar are you using? */
		return 0;
	}
	switch (month) {
		case 1:
		case 3:
		case 5:
		case 7:
		case 8:
		case 10:
		case 12:
			days = 31;
			break;
		case 2:
			days = 28;
			if (year % 4 == 0 &&
			    year % 100 != 0 &&
			    year % 400 == 0) {
			    	fprintf (stderr, "%s: leap year\n",
			    		__PRETTY_FUNCTION__);
				days++;
			}
			break;
		case 4:
		case 6:
		case 9:
		case 11:
			days = 30;
			break;
	}
	return days;
}


double GMST_degrees_from_datetime (DateTime *dt)
{
	double lst;
	double days;
	if (dt == NULL)
		return -1;
	days = epoch_j2000_days (dt);
	lst = 280.46061837 + 360.98564736629 * days;
	lst = normalize (lst, 360.0);
	return lst;
}

double LMST_degrees_from_datetime (DateTime *dt, double longitude)
{
	double lst;
	double days;
	if (dt == NULL)
		return -1;
	days = epoch_j2000_days (dt);
	lst = 280.46061837 + 360.98564736629 * days + longitude;
	lst = normalize (lst, 360.0);
	return lst;
}


double degrees_from_dms (int degrees, int arcminute, int arcsecond)
{
	double d;
	d = (double)degrees;
	if (degrees >= 0) {
		d += (double)arcminute / 60.0;
		d += (double)arcsecond / 3600.0;
	} else {
		d -= (double)arcminute / 60.0;
		d -= (double)arcsecond / 3600.0;
	}
	return d;
}

