/************************************************************************/
/* console.c - console functions source					*/
/************************************************************************/
/*
 *  (C)2004,2020 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <math.h>

#include <cutil/cutil.h>
#include <dbi/dbi.h>
#include <gtc/gtcdef.h>
#include <gtc/gtc.h>

#include "microfsm.h"
#include "skyobject.h"
#include "console.h"


/*
 Needed changes to GTC API:
 //short name (no whitespace)
C int gtc_module_get_name (GtcModule, char **name);
 //long name (whitespace acceptable)
+ int gtc_module_get_description (GtcModule, char **description);
*/


#define BUFFER_SIZE	8192


#define STDOUT_FD	1


enum console_signals_enum {
	CONSOLE_SIG_COMMANDSTR		= MICROFSM_SIG_USER,
};


typedef struct console_event_s	console_event_t;
struct console_event_s {
	microfsm_event_t	super;
	int			cmdlen;
	char			**cmditem;
	void			*userdata;
};



typedef struct _State State;
struct _State {
	microfsm_t	fsm;
	SList		*gtcmodulelist;
	GtcScope	*telescope;
	GtcObservatory	*observatory;
	char		*configfile;
	dbi_conn	dbconn;
//	SList		*skyviewobjectlist;
//	double		skyviewppu;
	unsigned long	flags;
//	SkyView		skyview;	
//	double		skyviewLST;
//	DateTime	skyviewDateTime;
	SkyObject	*target;
	/* BEGIN: Set from the config file */
	SList		*modulepathlist;
	char		*telescopemodulename;
	char		*observatorymodulename;
	char		*telescope_device;
	char		*observatory_device;
	char		*dbdriver;
	char		*dbhostname;
	char		*dbhostport;
	char		*dbname;
	char		*dbuser;
	char		*dbpassword;
	double		sitelongitude;
	double		sitelatitude;
	double		siteutcoffset;
	char		*capturedirectory;
	char		*captureprefix;
	double		captureinterval;
	int		capturecount;
	int		output_fd;

	int		state_id;

	/* END: Set from the config file */
};


// Static declarations
//GtcScope	*telescope;
//GtcObservatory	*observatory;
//SkyObject	*target;
//int		output_fd;
//char		*telescope_device;

State		state;



microfsm_result_t console_initial(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t console_config(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t console_module(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t console_observatory(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t console_telescope(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t console_object(microfsm_t *me, microfsm_event_t *event);


microfsm_result_t console_initial(microfsm_t *me, microfsm_event_t *event)
{
	switch (event->signal) {
		case MICROFSM_SIG_ENTRY:
			return MICROFSM_RET_HANDLED();
		case MICROFSM_SIG_EXIT:
			return MICROFSM_RET_HANDLED();
		case CONSOLE_SIG_COMMANDSTR:
			return MICROFSM_RET_HANDLED();
	}
	return MICROFSM_RET_IGNORED();
}



enum state_id_enum {
	STATE_ID_ROOT	= 0,
	STATE_ID_MODULE,
	STATE_ID_TELESCOPE,
	STATE_ID_OBSERVATORY,
	STATE_ID_OBJECT,
};


const char *state_str_list[] = {
	"",
	"module",
	"telescope",
	"observatory",
	"object",
};



int scan_degrees (char *string, double *degrees)
{
	int rc;
	int deg;
	int am;
	int as;
	double d;
	if (string == NULL || degrees == NULL)
		return GTC_RC_INVALIDARGUMENT;
fprintf (stderr, "%s: string=%s\n", __PRETTY_FUNCTION__, string);
	if (strstr (string, ".") != NULL) {
		d = atof (string);
fprintf (stderr, "%s: d=%f\n", __PRETTY_FUNCTION__, d);
	} else if (strstr (string, "d") != NULL) {
		rc = sscanf (string, "%3dd%2d'%2d''",
				&deg,
				&am,
				&as);
fprintf (stderr, "%s: sscanf rc=%d\n", __PRETTY_FUNCTION__, rc);
		if (rc != 3)
			return GTC_RC_INVALIDARGUMENT;
fprintf (stderr, "%s: deg=%d am=%d as=%d\n", __PRETTY_FUNCTION__, deg, am, as);
//		d = degrees_from_dms (deg,
//					am,
//					as);
fprintf (stderr, "%s: d=%f\n", __PRETTY_FUNCTION__, d);
	} else {
		return GTC_RC_INVALIDARGUMENT;
	}
	*degrees = d;
	return 0;
}	


int console_handle_command (char *commandline)
{
	int rc;
	int items;
	int fd;
	int bytes;
	char buffer[BUFFER_SIZE];
	char *str = NULL;
	char **cmditem = NULL;
	if (commandline == NULL)
		return EINVAL;

	if (strlen (commandline) == 0)
		return 0;
	items = 16;
	rc = strnsplit (commandline,
			strlen (commandline),
			' ',
			&cmditem,
			&items);
	if (rc != 0) {
		console_print_line ("strnsplit() failure.");
		return GTC_RC_INTERNALERROR;
	}
	/****************************************************************/
	/* quit 							*/
	/****************************************************************/
	if (!strcmp (cmditem[0], "quit")) {
		if (items == 1) {
			strarray_free (&cmditem);
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[1], "all")) {
			/* stop any motion */
			rc = gtc_scope_slew_stop (state.telescope);
			rc = gtc_scope_focus_stop (state.telescope);
			rc = gtc_scope_derotator_enable (state.telescope,
							FALSE);
			rc = gtc_scope_free (&state.telescope);
//			rc = gnuobserver_modules_unload();
//			rc = gnuobserver_config_items_free();
//			free (state.configfile);
			strarray_free (&cmditem);
			exit (0);
			return 0;
		}
	/****************************************************************/
	/* help 							*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "help")) {
		rc = console_handle_help_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* stop 							*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "stop")) {
		/* stop any motion */
		rc = gtc_scope_slew_stop (state.telescope);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
		rc = gtc_scope_focus_stop (state.telescope);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
		rc = gtc_scope_derotator_enable (state.telescope,
						FALSE);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
		console_print_line ("All scope motion stopped.");
	/****************************************************************/
	/* observatory 							*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "observatory")) {
		rc = console_handle_observatory_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* telescope	 						*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "telescope")) {
		rc = console_handle_telescope_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* gtcmodule	 						*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "gtcmodule")) {
		rc = console_handle_gtcmodule_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* console 							*/
	/****************************************************************/
//	} else if (!strcmp (cmditem[0], "console")) {
//		if (items == 1) {
//			strarray_free (&cmditem);
//			return GTC_RC_MISSINGARGUMENT;
//		}
//		if (!strcmp (cmditem[1], "clear")) {
//			gtk_text_buffer_get_start_iter (output_fd, &siter);
//			gtk_text_buffer_get_end_iter (output_fd, &eiter);
//			gtk_text_buffer_delete (output_fd, &siter, &eiter);
//		} else if (!strcmp (cmditem[1], "close")) {
//			state.flags |= FLAG_CONSOLETRANSIT;
//			gtk_widget_hide (state.windowConsole);
//			gtk_widget_destroy (state.windowConsole);
//			state.windowConsole = NULL;
//			state.flags -= (state.flags & FLAG_CONSOLETRANSIT);
//			state.flags -= (state.flags & FLAG_CONSOLEACTIVE);
//		} else if (!strcmp (cmditem[1], "save")) {
//			if (items != 3) {
//				strarray_free (&cmditem);
//				return GTC_RC_MISSINGARGUMENT;
//			}
//			gtk_text_buffer_get_start_iter (output_fd, &siter);
//			gtk_text_buffer_get_end_iter (output_fd, &eiter);
//			str = gtk_text_buffer_get_text (output_fd,
//							&siter,
//							&eiter,
//							FALSE);
//			fd = open (cmditem[2], O_WRONLY|O_CREAT, 0600);
//			if (fd == -1) {
//				rc = errno;
//				snprintf (buffer, BUFFER_SIZE, "Failed to create file '%s'.", cmditem[2]);
//				console_print_line (buffer);
//				strarray_free (&cmditem);
//				return rc;
//			}
//			bytes = strlen (str);
//			while (bytes) {
//				if (bytes > 4096)
//					rc = write (fd, str, 4096);
//				else
//					rc = write (fd, str, bytes);
//				if (rc == -1) {
//					rc = errno;
//					close (fd);
//					snprintf (buffer,
//						BUFFER_SIZE,
//						"Error writing to '%s': rc = %d(%s).",
//						cmditem[2],
//						errno,
//						strerror (errno));
//					console_print_line (buffer);
//					strarray_free (&cmditem);
//					return rc;
//				}
//				bytes -= rc;
//			}
//			rc = close (fd);
//			snprintf (buffer, BUFFER_SIZE, "Console text saved to '%s'.", cmditem[2]);
//			console_print_line (buffer);
//		} else {
//			strarray_free (&cmditem);
//			return GTC_RC_INVALIDCOMMAND;
//		}
	/****************************************************************/
	/* info 							*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "info")) {
		if (items == 1) {
			strarray_free (&cmditem);
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[1], "obscon")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"obscon version: %s",
					VERSION);
			console_print_line (buffer);
		} else {
			strarray_free (&cmditem);
			return GTC_RC_INVALIDCOMMAND;
		}
	/****************************************************************/
	/* config	 						*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "config")) {
		rc = console_handle_config_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* database	 						*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "database")) {
		rc = console_handle_database_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* capture	 						*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "capture")) {
		rc = console_handle_capture_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* object	 						*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "object")) {
		rc = console_handle_object_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* script	 						*/
	/****************************************************************/
	} else if (!strcmp (cmditem[0], "script")) {
		rc = console_handle_script_command (cmditem,
						items);
		if (rc != 0) {
			strarray_free (&cmditem);
			return rc;
		}
	/****************************************************************/
	/* invalid/unknown command					*/
	/****************************************************************/
	} else {
		strarray_free (&cmditem);
		console_print_line("Invalid command.");
		return GTC_RC_INVALIDCOMMAND;
	}		
	strarray_free (&cmditem);
	return 0;
}


int console_print_line (char *text)
{
	ssize_t	bytes;
	if (state.output_fd< 0 || text == NULL)
		return EINVAL;
	/* move cursor to the end of the buffer */

	bytes = write (state.output_fd, text, strlen(text));
	bytes = write (state.output_fd, "\n", 1);
	return 0;
}


int console_print_prompt ()
{
	ssize_t	bytes;
	char	buffer[BUFFER_SIZE];

//	GtkTextIter iter;
//	GtkTextView *gtv = NULL;
//	GtkTextMark *gtm = NULL;
	if (state.output_fd == -1)
		return EINVAL;

	// TODO fix this
	bytes = snprintf (buffer,
		BUFFER_SIZE,
		"\n%s> ",
		state_str_list[state.state_id]);
	bytes = write (state.output_fd, buffer, bytes);

	/* move cursor to the end of the buffer */
//	gtk_text_buffer_get_end_iter (output_fd, &iter);
//	gtk_text_buffer_place_cursor (output_fd, &iter);

//	gtk_text_buffer_insert_at_cursor (output_fd, "> ", 2);
//	gtv = GTK_TEXT_VIEW(lookup_widget (state.windowConsole,
//					"tvConsole"));
	/* scroll to the end of the buffer */
//	gtk_text_buffer_get_end_iter (output_fd, &iter);
//	gtm = gtk_text_buffer_create_mark (output_fd,
//					"end",
//					&iter,
//					TRUE);
//	gtk_text_view_scroll_to_mark (gtv,
//					gtm,
//					0,
//					TRUE,
//					0.0,
//					1.0);
	return 0;
}



int console_handle_telescope_command (char **cmditem,
			int items)
{
	int rc;
	int bytes;
	int idegree;
	int iarcminute;
	int iarcsecond;
	double dval;
	double degree;
	double arcminute;
	double arcsecond;
	GtcDate gdate;
	GtcTime gtime;
	GtcRa gra;
	GtcDec gdec;
	char buffer[BUFFER_SIZE];
	char *str = NULL;
	SList *list = NULL;
	GtcModule *gm = NULL;
	if (cmditem == NULL)
		return EINVAL;
	if (items == 1) {
		console_print_line ("Missing argument.");
		return GTC_RC_MISSINGARGUMENT;
	}
	if (!strcmp (cmditem[1], "disconnect")) {
		rc = gtc_scope_close (state.telescope);
		if (rc != 0) {
			bytes = snprintf (buffer,
					BUFFER_SIZE,
					"Failed to disconnect from telescope rc=%d (%s)",
					rc,
					gtc_get_strerror(rc));
			console_print_line (buffer);
			return rc;
		}
		console_print_line ("Telescope device disconnected.");
	} else if (!strcmp (cmditem[1], "connect")) {
		if (state.telescope_device == NULL) {
			console_print_line ("Telescope device not set.");
			return 0;
		}
		rc = gtc_scope_connect (state.telescope, state.telescope_device);
		if (rc != 0) {
			bytes = snprintf (buffer,
					BUFFER_SIZE,
					"Failed to connect to telescope rc=%d (%s)",
					rc,
					gtc_get_strerror(rc));
			console_print_line (buffer);
			return rc;
		}
		console_print_line ("Telescope device connected.");
	} else if (!strcmp (cmditem[1], "get")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "module")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Telescope module set to '%s'",
					state.telescopemodulename);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "device")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Telescope device set to '%s'",
					state.telescope_device);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "latitude")) {
			rc = gtc_scope_get_latitude (state.telescope, &dval);
			if (rc != 0)
				return rc;
			get_dms_from_degrees (dval,
						&degree,
						&arcminute,
						&arcsecond);
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current latitude: %02d deg %02d min %02d sec",
					(int)degree,
					(int)arcminute,
					(int)arcsecond);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "longitude")) {
			rc = gtc_scope_get_longitude (state.telescope, &dval);
			if (rc != 0)
				return rc;
			get_dms_from_degrees (dval,
						&degree,
						&arcminute,
						&arcsecond);
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current longitude: %02d deg %02d min %02d sec",
					(int)degree,
					(int)arcminute,
					(int)arcsecond);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "date")) {
			rc = gtc_scope_get_datetime (state.telescope,
							&gdate,
							&gtime);
			if (rc != 0)
				return rc;
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current scope date: %04d-%02d-%02d",
					gdate.year,
					gdate.month,
					gdate.day);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "time")) {
			rc = gtc_scope_get_datetime (state.telescope,
							&gdate,
							&gtime);
			if (rc != 0)
				return rc;
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current scope time: %02d:%02d:%02d",
					gtime.hour,
					gtime.minute,
					gtime.second);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "sidereal")) {
			rc = gtc_scope_get_sidereal_time (state.telescope,
							&gtime);
			if (rc != 0)
				return rc;
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current scope time: %02d:%02d:%02d",
					gtime.hour,
					gtime.minute,
					gtime.second);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "utcoffset")) {
			rc = gtc_scope_get_utc_offset (state.telescope,
							&dval);
			if (rc != 0)
				return rc;
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current scope UTC offset: %2.1f",
					dval);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "ra")) {
			rc = gtc_scope_get_current_ra (state.telescope,
							&gra);
			if (rc != 0)
				return rc;
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current scope RA: %02d:%02d:%02d",
					gra.hour,
					gra.minute,
					gra.second);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "dec")) {
			rc = gtc_scope_get_current_dec (state.telescope,
							&gdec);
			if (rc != 0)
				return rc;
			get_dms_from_degrees (gdec.deg,
						&degree,
						&arcminute,
						&arcsecond);
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current telescope DEC: %02d deg %02d min %02d sec",
					(int)degree,
					(int)arcminute,
					(int)arcsecond);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "model")) {
			rc = gtc_scope_get_model (state.telescope, &str);
			if (rc != 0)
				return rc;
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Telescope model: %s",
					str);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "firmware-version")) {
			rc = gtc_scope_get_firmware_version (state.telescope,
								&str);
			if (rc != 0)
				return rc;
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Telescope firmware version: %s",
					str);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "firmware-date")) {
			rc = gtc_scope_get_firmware_date (state.telescope,
								&str);
			if (rc != 0)
				return rc;
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Telescope firmware date: %s",
					str);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "target")) {
			if (items < 4) {
				console_print_line ("Missing argument.");
				return GTC_RC_MISSINGARGUMENT;
			}
			if (!strcmp (cmditem[3], "ra")) {
				rc = gtc_scope_get_target_ra (state.telescope,
								&gra);
				if (rc != 0)
					return rc;
				rc = snprintf (buffer,
						BUFFER_SIZE,
						"Current scope target RA: %02d:%02d:%02d",
						gra.hour,
						gra.minute,
						gra.second);
				console_print_line (buffer);
			} else if (!strcmp (cmditem[3], "dec")) {
				rc = gtc_scope_get_target_dec (state.telescope,
								&gdec);
				if (rc != 0)
					return rc;
			get_dms_from_degrees (gdec.deg,
						&degree,
						&arcminute,
						&arcsecond);
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Current telescope target DEC: %02d deg %02d min %02d sec",
					(int)degree,
					(int)arcminute,
					(int)arcsecond);
				console_print_line (buffer);
			} else {
				return GTC_RC_INVALIDCOMMAND;
			}
		} else {
			return GTC_RC_INVALIDCOMMAND;
		}
	} else if (!strcmp (cmditem[1], "set")) {
		if (items < 4) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "module")) {
			if (items != 4) {
				console_print_line ("Missing argument.");
				return GTC_RC_MISSINGARGUMENT;
			}
			list = state.gtcmodulelist;
			while (list != NULL) {
				gm = (GtcModule *)slist_get_data (list);
				rc = gtc_module_get_name (gm, &str);
				if (!strcmp (cmditem[3], str)) {
					if (state.telescopemodulename != NULL)
						free (state.telescopemodulename);
					state.telescopemodulename = strdup (str);
					rc = gtc_scope_free (&state.telescope);
					rc = gtc_scope_new (&state.telescope,
								gm);
					return rc;
				}
				list = slist_get_next (list);
			}
			return GTC_RC_INVALIDARGUMENT;
		} else if (!strcmp (cmditem[2], "device")) {
			if (items != 4) {
				console_print_line ("Missing argument.");
				return GTC_RC_MISSINGARGUMENT;
			}
			if (state.telescope_device != NULL)
				free (state.telescope_device);
			state.telescope_device = strdup (cmditem[3]);
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Telescope device set to '%s'",
					state.telescope_device);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "date")) {
			rc = sscanf (cmditem[3], "%2d/%2d/%2d",
					&gdate.month,
					&gdate.day,
					&gdate.year);

			if (rc != 3) {
				console_print_line ("Date format is invalid.");
				return GTC_RC_INVALIDARGUMENT;
			}
			if (gdate.month < 1 || gdate.month > 12)
				return GTC_RC_INVALIDARGUMENT;
			if (gdate.day < 1 || gdate.day > 31)
				return GTC_RC_INVALIDARGUMENT;
			if (gdate.year < 0 || gdate.year > 100)
				return GTC_RC_INVALIDARGUMENT;
			rc = gtc_scope_set_datetime (state.telescope,
						&gdate,
						NULL);
			if (rc != 0)
				return rc;
		} else if (!strcmp (cmditem[2], "time")) {
			rc = sscanf (cmditem[3], "%2d:%2d:%2d",
					&gtime.hour,
					&gtime.minute,
					&gtime.second);
			if (rc != 3)
				return GTC_RC_INVALIDARGUMENT;
			if (gtime.hour < 0 || gtime.hour > 23)
				return GTC_RC_INVALIDARGUMENT;
			if (gtime.minute < 0 || gtime.minute > 59)
				return GTC_RC_INVALIDARGUMENT;
			if (gtime.second < 0 || gtime.second > 59)
				return GTC_RC_INVALIDARGUMENT;
			rc = gtc_scope_set_datetime (state.telescope,
						NULL,
						&gtime);
			if (rc != 0)
				return rc;
		} else if (!strcmp (cmditem[2], "sidereal")) {
			rc = sscanf (cmditem[3], "%2d:%2d:%2d",
					&gtime.hour,
					&gtime.minute,
					&gtime.second);
			if (rc != 3)
				return GTC_RC_INVALIDARGUMENT;
			if (gtime.hour < 0 || gtime.hour > 23)
				return GTC_RC_INVALIDARGUMENT;
			if (gtime.minute < 0 || gtime.minute > 59)
				return GTC_RC_INVALIDARGUMENT;
			if (gtime.second < 0 || gtime.second > 59)
				return GTC_RC_INVALIDARGUMENT;
			rc = gtc_scope_set_sidereal_time (state.telescope,
						&gtime);
			if (rc != 0)
				return rc;
		} else if (!strcmp (cmditem[2], "utcoffset")) {
			dval = atof (cmditem[3]);
			if (dval < -12.0 || dval > 12.0)
				return GTC_RC_INVALIDARGUMENT;
			rc = gtc_scope_set_utc_offset (state.telescope,
						dval);
			if (rc != 0)
				return rc;
		} else if (!strcmp (cmditem[2], "longitude")) {
			rc = scan_degrees (cmditem[3], &dval);
			if (rc != 0)
				return GTC_RC_INVALIDARGUMENT;
			if (dval < -180.0 || dval > 180.0)
				return GTC_RC_INVALIDARGUMENT;
			rc = gtc_scope_set_longitude (state.telescope,
						dval);
			if (rc != 0)
				return rc;
		} else if (!strcmp (cmditem[2], "latitude")) {
			rc = scan_degrees (cmditem[3], &dval);
			if (rc != 0)
				return GTC_RC_INVALIDARGUMENT;
			if (dval < -90.0 || dval > 90.0)
				return GTC_RC_INVALIDARGUMENT;
			rc = gtc_scope_set_latitude (state.telescope,
						dval);
			if (rc != 0)
				return rc;
		} else if (!strcmp (cmditem[2], "target")) {
			if (items < 5) {
				console_print_line ("Missing argument.");
				return GTC_RC_MISSINGARGUMENT;
			}
			if (!strcmp (cmditem[3], "ra")) {
				rc = sscanf (cmditem[4], "%2d:%2d:%2d",
						&gra.hour,
						&gra.minute,
						&gra.second);
				if (rc != 3)
					return GTC_RC_INVALIDARGUMENT;
				if (gra.hour < 0 || gra.hour > 23)
					return GTC_RC_INVALIDARGUMENT;
				if (gra.minute < 0 || gra.minute > 59)
					return GTC_RC_INVALIDARGUMENT;
				if (gra.second < 0 || gra.second > 59)
					return GTC_RC_INVALIDARGUMENT;
				rc = gtc_scope_set_target_ra (state.telescope,
							&gra);
				if (rc != 0)
					return rc;
			} else if (!strcmp (cmditem[3], "dec")) {
				rc = scan_degrees (cmditem[4], &dval);
				if (rc != 0)
					return GTC_RC_INVALIDARGUMENT;
				if (dval < -90.0 || dval > 90.0)
					return GTC_RC_INVALIDARGUMENT;
				gdec.deg = dval; 
				rc = gtc_scope_set_target_dec (state.telescope,
							&gdec);
				if (rc != 0)
					return rc;
			} else {
				console_print_line ("Invalid argument.");
				return GTC_RC_INVALIDCOMMAND;
			}
		} else {
			console_print_line ("Invalid argument.");
			return GTC_RC_INVALIDCOMMAND;
		}
	} else if (!strcmp (cmditem[1], "target")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "goto")) {
			rc = gtc_scope_goto_target (state.telescope);
			if (rc != 0) {
				bytes = snprintf (buffer,
						BUFFER_SIZE,
						"Failed to issue got command to telescope rc=%d (%s)",
						rc,
						gtc_get_strerror(rc));
				console_print_line (buffer);
				return rc;
			}
		} else {
			console_print_line ("Invalid argument.");
			return GTC_RC_INVALIDCOMMAND;
		}
	} else if (!strcmp (cmditem[1], "align")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "auto")) {
			rc = gtc_scope_autoalign (state.telescope);
			if (rc != 0) {
				bytes = snprintf (buffer,
						BUFFER_SIZE,
						"Failed to issue alignment command to telescope rc=%d (%s)",
						rc,
						gtc_get_strerror(rc));
				console_print_line (buffer);
				return rc;
			}
		} else {
			console_print_line ("Invalid argument.");
			return GTC_RC_INVALIDCOMMAND;
		}
	} else {
		console_print_line ("Invalid command.");
		return GTC_RC_INVALIDCOMMAND;
	}
	return 0;
}



int console_handle_observatory_command (char **cmditem,
			int items)
{
	int rc;
	char buffer[BUFFER_SIZE];
	char *str = NULL;
	SList *list = NULL;
	GtcModule *gm = NULL;
	if (cmditem == NULL)
		return EINVAL;
	if (items == 1) {
		console_print_line ("Missing argument.");
		return GTC_RC_MISSINGARGUMENT;
	}
	if (!strcmp (cmditem[1], "disconnect")) {
		rc = gtc_observatory_close (state.observatory);
		if (rc != 0)
			return rc;
		console_print_line ("Observatory device disconnected.");
	} else if (!strcmp (cmditem[1], "connect")) {
		if (state.observatory_device == NULL) {
			console_print_line ("Observatory device not set.");
			return 0;
		}
		rc = gtc_observatory_connect (state.observatory,
					state.observatory_device);
		if (rc != 0)
			return rc;
		console_print_line ("Observatory device connected.");
	} else if (!strcmp (cmditem[1], "set")) {
		if (items < 4) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "module")) {
			if (items != 4) {
				console_print_line ("Missing argument.");
				return GTC_RC_MISSINGARGUMENT;
			}
			list = state.gtcmodulelist;
			while (list != NULL) {
				gm = (GtcModule *)slist_get_data (list);
				rc = gtc_module_get_name (gm, &str);
				if (!strcmp (cmditem[3], str)) {
					if (state.observatorymodulename != NULL)
						free (state.observatorymodulename);
					state.observatorymodulename = strdup (str);
					rc = gtc_observatory_free (&state.observatory);
					rc = gtc_observatory_new (&state.observatory,
								gm);
					return 0;
				}
				list = slist_get_next (list);
			}
			return GTC_RC_INVALIDARGUMENT;
		} else if (!strcmp (cmditem[2], "device")) {
			if (items != 4) {
				console_print_line ("Missing argument.");
				return GTC_RC_MISSINGARGUMENT;
			}
			if (state.observatory_device != NULL)
				free (state.observatory_device);
			state.observatory_device = strdup (cmditem[3]);
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Observatory device set to '%s'",
					state.observatory_device);
			console_print_line (buffer);
		} else {
			return GTC_RC_INVALIDCOMMAND;
		}
	} else if (!strcmp (cmditem[1], "get")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "module")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Observatory module set to '%s'",
					state.observatorymodulename);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "device")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Observatory device set to '%s'",
					state.observatory_device);
			console_print_line (buffer);
		} else {
			return GTC_RC_INVALIDCOMMAND;
		}
	} else {
		return GTC_RC_INVALIDCOMMAND;
	}
	return 0;
}



int console_handle_help_command (char **cmditem,
			int items)
{
	char buffer[BUFFER_SIZE];
	if (cmditem == NULL)
		return EINVAL;
	if (items < 2) {
		console_print_line ("help [command]                        Display help for the given command.");
		console_print_line ("quit [subcmd]                         Halt scope and quit state.");
		console_print_line ("stop                                  Halt all scope motion.");
		console_print_line ("telescope [subcmd]                    Perform telescope operations.");
		console_print_line ("observatory [subcmd]                  Perform observatory operations.");
		console_print_line ("gtcmodule [subcmd]                    Perform GTC module operations.");
		console_print_line ("console [subcmd]                      Alter console attributes.");
		console_print_line ("info [subcmd]                         Display specific item information.");
		console_print_line ("config [subcmd]                       Perform configuration operations.");
		console_print_line ("database [subcmd]                     Perform database operations.");
		console_print_line ("capture [subcmd]                      Perform capture operations.");
		console_print_line ("object [subcmd]                       Perform object operations.");
		console_print_line ("script [subcmd]                       Perform script operations.");
		return 0;
	}
//	if (!strcmp (cmditem[1], "console")) {
//		console_print_line ("console clear                        Clear the console contents.");
//		console_print_line ("console close                        Close the console window.");
//		console_print_line ("console save {filename}              Save the console's text to a file.");
	if (!strcmp (cmditem[1], "script")) {
		console_print_line ("script execute {filename}            Execute commands from the given script file.");
	} else if (!strcmp (cmditem[1], "quit")) {
		console_print_line ("quit all                             Stop motion and quit state.");
	} else if (!strcmp (cmditem[1], "observatory")) {
		if (items == 2) {
			console_print_line ("observatory connect                  Connect to the observatory device.");
			console_print_line ("observatory disconnect               Disconnect from the observatory device.");
			console_print_line ("observatory set {parameter} {value}  Alter an observatory parameter.");
			console_print_line ("observatory get {parameter}          Display an observatory parameter.");
		} else if (items == 3) {
			if (!strcmp (cmditem[2], "set")) {
				console_print_line ("set module {name}                    Set observatory module name.");
				console_print_line ("set device {path}                    Set observatory device path.");
			} else if (!strcmp (cmditem[2], "get")) {
				console_print_line ("get module                           Get observatory module name.");
				console_print_line ("get device                           Get observatory device path.");
			}
		}
	} else if (!strcmp (cmditem[1], "telescope")) {
		if (items == 2) {
			console_print_line ("telescope connect                    Connect to the telescope device.");
			console_print_line ("telescope disconnect                 Disconnect from the telescope device.");
			console_print_line ("telescope set {parameter} {value}    Alter a telescope parameter.");
			console_print_line ("telescope get {parameter}            Display a telescope parameter.");
			console_print_line ("telescope target goto                Goto the telescope's current target location.");
			console_print_line ("telescope align auto                 Perform automatic telescope alignment.");
		} else if (items == 3) {
			if (!strcmp (cmditem[2], "get")) {
				console_print_line ("get module                           Get telescope module name.");
				console_print_line ("get device                           Get telescope device path.");
				console_print_line ("get date                             Get telescope date.");
				console_print_line ("get time                             Get telescope time.");
				console_print_line ("get sidereal                         Get telescope sidereal time.");
				console_print_line ("get utcoffset                        Get telescope UTC ofset.");
				console_print_line ("get longitude                        Get telescope longitude.");
				console_print_line ("get latitude                         Get telescope latitude.");
				console_print_line ("get target ra                        Get telescope target RA.");
				console_print_line ("get target dec                       Get telescope target declination.");
				console_print_line ("get model                            Get telescope model information.");
				console_print_line ("get firmware-version                 Get telescope firmware version.");
				console_print_line ("get firmware-date                    Get telescope firmware date.");
			} else if (!strcmp (cmditem[2], "set")) {
				console_print_line ("set module {name}                    Set telescope module name.");
				console_print_line ("set device {path}                    Set telescope device path.");
				console_print_line ("set date {mm/dd/yy}                  Set telescope date.");
				console_print_line ("set time {hh:mm:ss}                  Set telescope time.");
				console_print_line ("set sidereal {hh:mm:ss}              Set telescope sidereal time.");
				console_print_line ("set utcoffset {sHH.H}                Set telescope UTC ofset.");
				console_print_line ("set longitude {sDDD.DDD|DDdMM'SS''}  Set telescope longitude.");
				console_print_line ("set latitude  {sDDD.DDD|DDdMM'SS''}  Set telescope latitude.");
				console_print_line ("set target ra {hh:mm:ss}             Set telescope target RA.");
				console_print_line ("set target dec {sDDD.DDD|DDdMM'SS''} Set telescope target declination.");
			}
		}
	} else if (!strcmp (cmditem[1], "info")) {
		console_print_line ("info module                          Display GTC module information.");
		console_print_line ("info telescope                       Display GTC telescope information.");
		console_print_line ("info observatory                     Display GTC observatory information.");
		console_print_line ("info obscon                          Display obscon information.");
	} else if (!strcmp (cmditem[1], "stop")) {
		console_print_line ("stop                                 Halt all telescope motion.");
	} else if (!strcmp (cmditem[1], "gtcmodule")) {
		console_print_line ("load {modulepath}                    Load the given module.");
		console_print_line ("unload {modulename}                  Unload the given module.");
		console_print_line ("info                                 Display loaded module information.");
	} else if (!strcmp (cmditem[1], "config")) {
		console_print_line ("save {configfile}                    Save the current configuration.");
	} else if (!strcmp (cmditem[1], "database")) {
		if (items == 2) {
			console_print_line ("database connect                  Connect to the database server.");
			console_print_line ("database disconnect               Disconnect from the database server.");
			console_print_line ("database set {parameter} {value}  Alter a database parameter.");
			console_print_line ("database get {parameter}          Display a database parameter.");
		} else if (items == 3) {
			if (!strcmp (cmditem[2], "set")) {
				console_print_line ("set driver {driver}                  Set database driver.");
				console_print_line ("set hostname {hostname}              Set database hostname.");
				console_print_line ("set hostport {port}                  Set database hostport.");
				console_print_line ("set dbname {dbname}                  Set database name.");
				console_print_line ("set username {username}              Set database username.");
				console_print_line ("set password {password}              Set database password.");
			} else if (!strcmp (cmditem[2], "get")) {
				console_print_line ("get driver                           Get database driver.");
				console_print_line ("get hostname                         Get database hostname.");
				console_print_line ("get hostport                         Get database hostport.");
				console_print_line ("get dbname                           Get database name.");
				console_print_line ("get username                         Get database username.");
				console_print_line ("get password                         Get database password.");
			}
		}
	} else if (!strcmp (cmditem[1], "capture")) {
		if (items == 2) {
			console_print_line ("capture set {parameter} {value}    Alter a capture parameter.");
			console_print_line ("capture get {parameter}            Display a capture parameter.");
			console_print_line ("capture start                      Start the capture using the current parameters.");
			console_print_line ("capture stop                       Stop the currently executing capture.");
		} else if (items == 3) {
			if (!strcmp (cmditem[2], "get")) {
				console_print_line ("get directory                        Get capture directory.");
				console_print_line ("get prefix                           Get capture filename prefix.");
				console_print_line ("get interval                         Get capture interval.");
				console_print_line ("get count                            Get capture count.");
			} else if (!strcmp (cmditem[2], "set")) {
				console_print_line ("set directory {path}                 Set capture directory.");
				console_print_line ("set prefix {prefix}                  Set capture filename prefix.");
				console_print_line ("set interval {seconds}               Set capture interval delay (SSS:ss seconds).");
				console_print_line ("set count {count}                    Set capture count.");
			}
		}
	} else if (!strcmp (cmditem[1], "object")) {
		if (items == 2) {
			console_print_line ("object search {parameter=value}    Search the object database.");
			console_print_line ("object select {id}                 Select an object from the search results.");
			console_print_line ("object goto                        Slew the telescope to the selected object.");
		}
	} else {
	snprintf (buffer, BUFFER_SIZE, "Help: command '%s' invalid.", cmditem[1]);
	console_print_line (buffer);
	}
	return 0;
}


int console_handle_gtcmodule_command (char **cmditem,
			int items)
{
	int rc;
	int count;
	char buffer[BUFFER_SIZE];
	char *name = NULL;
	char *version = NULL;
	char *desc = NULL;
	char *str = NULL;
	SList *list = NULL;
	GtcModule *gm = NULL;
	GtcModule *gmn = NULL;
	if (cmditem == NULL)
		return EINVAL;
	if (items == 1) {
		console_print_line ("Missing argument.");
		return GTC_RC_MISSINGARGUMENT;
	}
	if (!strcmp (cmditem[1], "load")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		rc = gtc_module_load (&gmn, cmditem[2]);
		if (rc != 0) {
			snprintf(buffer,
				BUFFER_SIZE,
				"Error loading module rc=%d %s",
				rc,
				gtc_get_strerror(rc));
			console_print_line (buffer);
			return rc;
		}
		rc = gtc_module_get_name (gmn, &name);
		list = state.gtcmodulelist;
		while (list != NULL) {
			gm = (GtcModule *)slist_get_data (list);
			rc = gtc_module_get_name (gm, &str);
			if (!strcmp (name, str)) {
				rc = gtc_module_unload (&gmn);
				console_print_line ("Module already loaded.");
				return GTC_RC_MODULEINUSE;
			}
			list = slist_get_next (list);
		}
		rc = slist_item_add (&state.gtcmodulelist, gmn);
		console_print_line ("GTC module loaded.");
	} else if (!strcmp (cmditem[1], "unload")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		list = state.gtcmodulelist;
		while (list != NULL) {
			gm = (GtcModule *)slist_get_data (list);
			rc = gtc_module_get_name (gm, &name);
			if (!strcmp (cmditem[2], name)) {
				rc = gtc_module_get_usecount (gm, &count);
				if (rc != 0) {
					console_print_line ("Error getting module use count.");
					return rc;
				}
				if (count != 0) {
					console_print_line ("Module is in use and cannot be unloaded at this time.");
					return GTC_RC_MODULEINUSE;
				}
				rc = slist_item_remove (&state.gtcmodulelist,
							gm);
				rc = gtc_module_unload (&gm);
				console_print_line ("GTC module unloaded.");
				return 0;
			}
			list = slist_get_next (list);
		}
		return GTC_RC_INVALIDARGUMENT;
	} else if (!strcmp (cmditem[1], "info")) {
		console_print_line ("GTC Module Information");
		console_print_line ("Name             Version          Description");
		console_print_line (
		"---------------------------------------"
		"---------------------------------------"
		"---------------------------------------");
		list = state.gtcmodulelist;
		while (list != NULL) {
			gm = (GtcModule *)slist_get_data (list);
			rc = gtc_module_get_name (gm, &name);
			rc = gtc_module_get_description (gm, &desc);
			rc = gtc_module_get_version (gm, &version);
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"%-16s %-16s  %s",
					name,
					version,
					desc);
			console_print_line (buffer);
			list = slist_get_next (list);
		}
	} else {
		return GTC_RC_INVALIDCOMMAND;
	}
	return 0;
}


int console_handle_config_command (char **cmditem,
			int items)
{
	int rc;
	char buffer[BUFFER_SIZE];
	char *str = NULL;
	if (cmditem == NULL)
		return EINVAL;
	if (items == 1) {
		console_print_line ("Missing argument.");
		return GTC_RC_MISSINGARGUMENT;
	}
	if (!strcmp (cmditem[1], "save")) {
		if (items == 3) {
			str = cmditem[2];
		} else {
			str = state.configfile;
		}
//		rc = gnuobserver_save_config (str);
//		if (rc != 0)
//			return rc;
		rc = snprintf (buffer,
				BUFFER_SIZE,
				"Configuration saved to '%s'",
				str);
		console_print_line (buffer);
	} else {
		return GTC_RC_INVALIDCOMMAND;
	}
	return 0;
}


int console_handle_database_command (char **cmditem,
			int items)
{
	int rc;
	char buffer[BUFFER_SIZE];
	if (cmditem == NULL)
		return EINVAL;
	if (items == 1) {
		console_print_line ("Missing argument.");
		return GTC_RC_MISSINGARGUMENT;
	}
	if (!strcmp (cmditem[1], "set")) {
		if (items < 4) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "hostname")) {
			if (state.dbhostname != NULL)
				free (state.dbhostname);
			state.dbhostname = strdup (cmditem[3]);
		} else if (!strcmp (cmditem[2], "hostport")) {
			if (state.dbhostport != NULL)
				free (state.dbhostport);
			state.dbhostport = strdup (cmditem[3]);
		} else if (!strcmp (cmditem[2], "driver")) {
			if (state.dbdriver != NULL)
				free (state.dbdriver);
			state.dbdriver = strdup (cmditem[3]);
		} else if (!strcmp (cmditem[2], "dbname")) {
			if (state.dbname != NULL)
				free (state.dbname);
			state.dbname = strdup (cmditem[3]);
		} else if (!strcmp (cmditem[2], "username")) {
			if (state.dbuser != NULL)
				free (state.dbuser);
			state.dbuser = strdup (cmditem[3]);
		} else if (!strcmp (cmditem[2], "password")) {
			if (state.dbpassword != NULL)
				free (state.dbpassword);
			state.dbpassword = strdup (cmditem[3]);
		} else {
			return GTC_RC_INVALIDCOMMAND;
		}
	} else if (!strcmp (cmditem[1], "get")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "hostname")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Database hostname is set to '%s'",
					state.dbhostname);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "hostport")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Database hostport is set to '%s'",
					state.dbhostport);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "driver")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Database driver is set to '%s'",
					state.dbdriver);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "dbname")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Database name is set to '%s'",
					state.dbname);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "username")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Database username is set to '%s'",
					state.dbuser);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "password")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Database password is set to '%s'",
					state.dbpassword);
			console_print_line (buffer);
		} else {
			return GTC_RC_INVALIDCOMMAND;
		}
	} else if (!strcmp (cmditem[1], "connect")) {
		if (state.dbconn != NULL)
			return GTC_RC_DEVICEBUSY;
		rc = dbi_initialize (NULL);
		if (rc == -1)
			return GTC_RC_INTERNALERROR;
		state.dbconn = dbi_conn_new (state.dbdriver);
		if (state.dbconn == NULL) {
			dbi_shutdown();
			return GTC_RC_INTERNALERROR;
		}
		rc = dbi_conn_set_option (state.dbconn, "host", state.dbhostname);
		rc = dbi_conn_set_option (state.dbconn, "port", state.dbhostport);
		rc = dbi_conn_set_option (state.dbconn, "username", state.dbuser);
		rc = dbi_conn_set_option (state.dbconn, "password", state.dbpassword);
		rc = dbi_conn_connect (state.dbconn);
		if (rc != 0) {
			dbi_conn_close (state.dbconn);
			state.dbconn = NULL;
			dbi_shutdown();
			return GTC_RC_INTERNALERROR;
		}
		rc = dbi_conn_select_db (state.dbconn, state.dbname);
		if (rc != 0) {
			dbi_conn_close (state.dbconn);
			state.dbconn = NULL;
			dbi_shutdown();
			return GTC_RC_INTERNALERROR;
		}
		console_print_line ("Database connected.");
	} else if (!strcmp (cmditem[1], "disconnect")) {
		if (state.dbconn == NULL)
			return GTC_RC_DEVICENOTPRESENT;
		dbi_conn_close (state.dbconn);
		state.dbconn = NULL;
		dbi_shutdown();
		console_print_line ("Database disconnected.");
	} else {
		return GTC_RC_INVALIDCOMMAND;
	}
	return 0;
}


int console_handle_capture_command (char **cmditem,
			int items)
{
	int rc;
	char buffer[BUFFER_SIZE];
	if (cmditem == NULL)
		return EINVAL;
	if (items == 1) {
		console_print_line ("Missing argument.");
		return GTC_RC_MISSINGARGUMENT;
	}
	if (!strcmp (cmditem[1], "set")) {
		if (items < 4) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "directory")) {
			if (state.capturedirectory != NULL)
				free (state.capturedirectory);
			state.capturedirectory = strdup (cmditem[3]);
		} else if (!strcmp (cmditem[2], "prefix")) {
			if (state.captureprefix != NULL)
				free (state.captureprefix);
			state.captureprefix = strdup (cmditem[3]);
		} else if (!strcmp (cmditem[2], "interval")) {
			state.captureinterval = atof (cmditem[3]);
		} else if (!strcmp (cmditem[2], "count")) {
			state.capturecount = atoi (cmditem[3]);
		} else {
			return GTC_RC_INVALIDCOMMAND;
		}
	} else if (!strcmp (cmditem[1], "get")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		if (!strcmp (cmditem[2], "directory")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Capture directory is set to '%s'",
					state.capturedirectory);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "prefix")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Capture prefix is set to '%s'",
					state.captureprefix);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "interval")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Capture interval is set to '%3.2f'",
					state.captureinterval);
			console_print_line (buffer);
		} else if (!strcmp (cmditem[2], "count")) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"Capture count is set to '%d'",
					state.capturecount);
			console_print_line (buffer);
		} else {
			return GTC_RC_INVALIDCOMMAND;
		}
	} else if (!strcmp (cmditem[1], "start")) {
	} else if (!strcmp (cmditem[1], "stop")) {
	} else {
		return GTC_RC_INVALIDCOMMAND;
	}
	return 0;
}



int console_handle_object_command (char **cmditem,
			int items)
{
	int rc;
	char buffer[BUFFER_SIZE];
	if (cmditem == NULL)
		return EINVAL;
	if (items == 1) {
		console_print_line ("Missing argument.");
		return GTC_RC_MISSINGARGUMENT;
	}
	if (!strcmp (cmditem[1], "search")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
	} else if (!strcmp (cmditem[1], "select")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
	} else if (!strcmp (cmditem[1], "goto")) {

		if (state.target == NULL)
			return 0;
		rc = snprintf (buffer,
				BUFFER_SIZE,
				"Slewing to: %s...",
				state.target->name);
		console_print_line (buffer);
		if (state.telescope == NULL) {
			console_print_line ("GOTO: telescope not connected.");
			return EINVAL;
		}
		rc = gtc_scope_set_target_ra (state.telescope,
					&state.target->ra);
		if (rc != 0) {
			console_print_line ("GOTO: target RA error");
			return rc;
		}
		rc = gtc_scope_set_target_dec (state.telescope,
					&state.target->dec);
		if (rc != 0) {
			console_print_line ("GOTO: target DEC error");
			return rc;
		}
		rc = gtc_scope_goto_target (state.telescope);
		if (rc != 0) {
			console_print_line ("GOTO: target goto error");
			return rc;
		}
	} else {
		return GTC_RC_INVALIDCOMMAND;
	}
	return 0;
}


int console_handle_script_command (char **cmditem,
			int items)
{
	int rc;
	int bytes;
	int lineno;
	char line[BUFFER_SIZE];
	char buffer[BUFFER_SIZE];
	char *p;
	FILE *fd = NULL;
	if (cmditem == NULL)
		return EINVAL;
	if (items == 1) {
		console_print_line ("Missing argument.");
		return GTC_RC_MISSINGARGUMENT;
	}
	if (!strcmp (cmditem[1], "execute")) {
		if (items < 3) {
			console_print_line ("Missing argument.");
			return GTC_RC_MISSINGARGUMENT;
		}
		fd = fopen(cmditem[2], "r");
		if (fd == NULL) {
			rc = errno;
			bytes = snprintf (buffer,
					BUFFER_SIZE,
					"Failed to open script %s rc=%d (%s)",
					cmditem[2],
					rc,
					gtc_get_strerror(rc));
			console_print_line (buffer);
			return rc;
		}
		lineno = 0;
		while (1) {
			p = fgets(line, BUFFER_SIZE, fd);
			if (p == NULL) {
				rc = errno;
				if (errno != 0) {
					bytes = snprintf (buffer,
							BUFFER_SIZE,
							"Error reading script %s rc=%d (%s)",
							cmditem[2],
							rc,
							gtc_get_strerror(rc));
					console_print_line (buffer);
					bytes = fclose(fd);
					return rc;
				} else {
					break; // normal EOF
				}
			}
			lineno++;
			bytes = strlen(line);
			if (bytes < 3) {
				// skip empty lines
				continue;
			}
			if (line[0] == '#') {
				// skip comment lines
				continue;
			}
			// remove the newline
			if (line[bytes-1] == '\n')
				line[bytes-1] = 0;

			// Execute command
			bytes = snprintf (buffer,
					BUFFER_SIZE,
					"Executing> %s",
					line);
			console_print_line (buffer);
			rc = console_handle_command(line);
			if (rc != 0) {
				bytes = fclose(fd);
				bytes = snprintf (buffer,
						BUFFER_SIZE,
						"Aborting script execution at line %d.",
						lineno);
				console_print_line (buffer);
				// attempt to stop any motion that may have been started
				bytes = gtc_scope_slew_stop(state.telescope);
				bytes = gtc_scope_focus_stop(state.telescope);
				return rc;
			}
			console_print_line ("");
		}
		rc = fclose(fd);
		console_print_line ("Script executed successfull.");
	} else {
		return GTC_RC_INVALIDCOMMAND;
	}
	return 0;
}



int console_initialize(int out_fd)
{
	microfsm_result_t	rc;

	state.telescope = NULL;
	state.observatory = NULL;
	state.target = NULL;
	state.output_fd = out_fd;
	state.state_id = 0;

	rc = microfsm_init(&state.fsm, &console_initial);

	return 0;
}


